/******************************************************************************
 * This demo for the MAXQ2000 uses an LCD and a thermistor to display the
 * current temperature.  It also uses pushbuttons to control the LCD 
 * contrast.
 *
 * Features demostrated in this demo:
 *   1) LCD controller
 *   2) Timers
 *   3) External Interrupts
 *   4) SPI communication
 *   5) Analog to digital conversion with external hardware (MAX1407)
 *   6) Serial communication
 *
 * Functional overview:
 *   The timer is set to generate an interrupt at a regular interval.  On an
 *   interrupt, the part performs a temperature reading, updates the LCD, and
 *   outputs the current temperature to the serial port.  The units will also
 *   toggle between Celcius and Farhenheit after 20 interrupts.
 *
 *   For simplicity, the temperature reading in this demo application is not 
 *   entirely accurate.  It is reasonably accurate at and around room 
 *   temperature.  
 *
 * Parts required:
 *   (1) MAXQ2000 EV-KIT Rev B with attached LCD (+ JTAG board, power, serial cable)
 *   (1) 10K Resistor
 *   (1) RL1005-5744-103-SA Thermistor
 *       [http://rocky.digikey.com/scripts/ProductInfo.dll?Site=US&V=45&M=RL1005-5744-103-SA]
 *
 * Wiring requirements:
 *
 *  The thermistor should be wired as follows:
 *
 *       10K Ohm between pin 1 of J7 and GND (1407:IN0 & GND)
 *       Thermistor between pin 1 and pin 7 of J7 (1407:IN0 & 1407:OUT2)
 *       Short between pin 7 and 8 of J7 (1407:OUT2 & 1407:FB2)
 *
 *  These switches should be turned on:
 *
 *       SW1.7               : Enable Serial 0 transmit
 *       SW6.2, SW6.5, SW6.8 : Enable push button interrupts on P5.2 and P7.1
 *       SW3.0 to SW3.7      : All on to enable comm with the MAX1407
 *
 *  Once everything has been wired, you can run the demo from within IAR's
 *  Embedded Workbench IDE.  After the demo is running, you should see the
 *  current temperature displayed on the LCD in degrees Celcius.
 *
 *****************************************************************************/
 
#include <intrinsics.h>
#include <iomaxq200x.h>

#define USE_SERIAL 1

#if (USE_SERIAL)
#include <stdio.h>
#endif

/*
 * Command codes for the MAX1407
 */
#define WR_ADC       0x080                    // 10   00000    0
#define WR_MUX       0x082                    // 10   00001    0
#define WR_DAC2      0x08A                    // 10   00101    0
#define WR_RUN       0x0BA                    // 10   11101    0

#define RD_ADC       0x0C0                    // 11   00000    0
#define RD_DATA      0x0C4                    // 11   00010    0
#define RD_DAC2      0x0CA                    // 11   00101    0
#define RD_STAT      0x0CC                    // 11   00110    0

/* 
 * Some definitions for LCD patterns.  For a full LCD digit, just index
 * into the PATTERNS array and write that value to the appropriate LCD register.
 *
 * Compatible with the MAXQ2000 EVKit Rev. B
 */
#define LCD0_PATTERN_C    0x039
#define LCD0_PATTERN_F    0x071

#define LCD_PATTERN_0     0x03F
#define LCD_PATTERN_1     0x006
#define LCD_PATTERN_2     0x05B
#define LCD_PATTERN_3     0x04F
#define LCD_PATTERN_4     0x066
#define LCD_PATTERN_5     0x06D
#define LCD_PATTERN_6     0x07D
#define LCD_PATTERN_7     0x007
#define LCD_PATTERN_8     0x07F
#define LCD_PATTERN_9     0x067

int PATTERNS[] = { LCD_PATTERN_0, LCD_PATTERN_1, LCD_PATTERN_2, LCD_PATTERN_3, LCD_PATTERN_4, 
                   LCD_PATTERN_5, LCD_PATTERN_6, LCD_PATTERN_7, LCD_PATTERN_8, LCD_PATTERN_9 };

/*
 * Flag - are we displaying celsius or farenheit temperatures?
 */
unsigned char celsius = 0;

/*
 * Counts how many temperature readings we have made.  After 20, we will change
 * our display mode (celsius <--> farenheit)
 */
int count = 0;

/*
 * Debug variables
 */
int pbcount = 0;
int last_adc = 0;
int last_temp = 0;
int lastEIFvalue;


/******************************************************************************
 * Performs an SPI transaction...output the value of 'spib' onto the
 * SPI bus, and return whatever was read.
 */
unsigned int sendSPI(unsigned int spib)
{
  SPIB = spib;                  // Load the data to send
  while(SPICN_bit.STBY);        // Loop until the data has been sent.
  SPICN_bit.SPIC = 0;           // Clear the SPI tansfer complete flag.
  return SPIB;  
}

/******************************************************************************
 * Gets an A->D reading from the MAX1407.  Communicates over SPI.
 * 16-bits are returned from the 1407.
 */
unsigned int getADCReading()
{
  unsigned int spiData;
  
  sendSPI(RD_ADC);                    // Read the 1407's ADC register.
  spiData = sendSPI(0x0FF);
  spiData |= 0x01;                    // Set the start conversion bit.
  sendSPI(WR_ADC);                    // Write the new value back into
  sendSPI(spiData);                   //   the 1407 register.
  
  do {
    sendSPI(RD_STAT);                 // Read the 1407's Status register.
    spiData = sendSPI(0x0FF);
  } while((spiData & 0x02) == 0x00);  // Bit 1 indicates AtoD conversion complete.
  
  sendSPI(RD_DATA);                   // Send the command to read the 1407's Data register.
  SPICF = 0x04;                       // Set up SPI to 16-bit mode.
  spiData = sendSPI(0x0FFFF);         // Read the Data register.
  SPICF = 0x00;                       // Put SPI  back into 8-bit mode.
  
  return spiData;
}

/******************************************************************************
 * Initializes the MAX1407.  Configures SPI for rising edge, 8-bit mode, master
 * mode.  3 readings must be taken before the 1407's results become valid.
 */
void init1407()
{
  unsigned int spiData;
  int i;
  
  PD5 |= 0x070;             // Set SPI output pins (CS, SCLK, DOUT) as output.
  PD5 &= ~0x080;            // Set SPI input pin (DIN) as input.

  SPICK = 0x10;             // Configure SPI for rising edge, sample input
  SPICF = 0x00;             //   on inactive edge, 8 bit, divide by 16.
  SPICN_bit.MSTM = 1;       // Set Q2000 as the master.
  SPICN_bit.SPIEN = 1;      // Enable SPI.

  P05 &= ~0x10;             // Set /CS low.

  sendSPI(WR_RUN);          // Make sure 1407 is up and running by putting
                            //   it into RUN mode.

  sendSPI(WR_ADC);          // Set up the 1407 ADC to sample at 30Hz, unbuffered,
  sendSPI(0x10);            //   in unipolar mode.  Also sets the PGA to 1.

  sendSPI(WR_MUX);          // Set up the 1407 MUXs.  The positive MUX is set to
  sendSPI(0x2C);            //   REF and the negative MUX is set to IN0.

  sendSPI(RD_ADC);          // Read the 1407's ADC register.
  spiData = sendSPI(0x0FF);
  spiData |= 0x81;          // Set the calibration & start conversion bit.
  sendSPI(WR_ADC);          // Write the new value back into
  sendSPI(spiData);         //   the 1407 register.
  do {
    sendSPI(RD_STAT);                 // Read the 1407's Status register.
    spiData = sendSPI(0x0FF);
  } while((spiData & 0x02) == 0x00);  // Bit 1 indicates AtoD conversion complete.

  sendSPI(WR_DAC2);                   // Set up the DAC2 to output 1.25V.  (FB2
                                      //   must be directly connected to OUT2).
  SPICF_bit.CHR = 1;                  // Set SPI to 16-bit mode.
  sendSPI(0x0FFC0);

  SPICF_bit.CHR = 0;                  // Put SPI back into 8-bit mode.

  for(i = 0; i < 3; i++)              // You must do three ADC readings on the 1407
  {                                   // before the results become valid.
    getADCReading();
  }
}
/******************************************************************************
 * Converts an A->D reading received from the MAX1407 into a celsius 
 * temperature approximation.
 */
int convertToTemp(unsigned int adc)
{
  double temp = -0.00135477 * (double)adc + 69.17;
  return (int)temp;
}

/******************************************************************************
 * Write a character to serial 0.  Interrupts are not used for serial 
 * transmission.  We are just bit-banging.  Returns the input value 'ch'.
 */
#if (USE_SERIAL)
int putchar(int ch)
{
  while(SCON0_bit.TI == 0);       // Wait until we can send.
  SCON0_bit.TI = 0;               // Clear the sent flag.
  SBUF0 = ch;                     // Send the char.
  return ch;
}

/******************************************************************************
 * Initialize the serial port to mode 1, baud rate 115200.  The equation for
 * the phase register setting is explained in the code below.
 */
void initSerial()
{
  SCON0_bit.SM1 = 1;              // Set to Mode 1.
  SCON0_bit.REN = 0;              // Disable receives.  Ain't gonna do anything with them anyway.
  SMD0_bit.SMOD = 1;              // Set baud rate to 16 times the baud clock.
  
  /*
   * Our baud rate equation is:
   *      (PR * FREQ)
   *  -------------------     <-- divided by
   *  (2^23 / 2^(SMOD*2))
   *
   * MAXQ2000 EV Kit (Rev B) comes with a 13.5 MHz crystal.  We want 115200.  We are using SMOD=1.
   *   Baud = (PR * 13500000)/(2^23/2^2) = PR * (13500000 / 2097152) = 115200
   *   PR = 115200 * 2097152 / 13500000 ~= 17896 = 45E8h
   */
  
//  PR0 = 0x45E8;                   // Set the phase for 115200 with a 13.5MHz crystal.
  asm("move M2[9], 0x45E8");
  
  SCON0_bit.TI = 0;               // Clear the transmit flag.
  SBUF0 = 0x0D;                   // Send a carriage return to start of communication.
}
#endif

/******************************************************************************
 * Lights up the solitary 'one' segment on our LCD panel.
 */
show100s()
{
  LCD4 = 0x40;
}

/******************************************************************************
 * Returns the value that will need to be placed in one of LCD0-LCD3 to display 
 * a digit 0-9.  No bounds checking is done here.  If you ask for a digit other
 * than 0-9, you will get a bogus display.
 */
int getLCDDigit(int digit)
{
    return PATTERNS[digit];
}

/******************************************************************************
 * Take a celsius temperature and output it to the LCD drive.  Figure out if
 * we want to display a celsius or farenheit value also.  Change the output
 * units appropriately.
 */
void showTemp(int temp)
{
#if (USE_SERIAL)
  printf("Temperature is ");
#endif

  // Clear the display of everything but the temperature units (F or C).
  LCD0 = ( (celsius == 0) ? LCD0_PATTERN_F : LCD0_PATTERN_C);
  // no degrees mark on this LCD board (we could do the upper half of an '8' as the degree sign, but it looks odd)
  LCD1 = 0;
  // clear the digits
  LCD2 = LCD3 = LCD4 = 0;

  if(celsius == 0)
  {
    // Convert to Fahrenheit.
    temp = (int)(((double)temp) * 9.0 / 5.0 + 32.0);
  }

  if (temp > 199)
  {
      temp = temp % 200;
  }

  if(temp > 99)
  {
    show100s();                         // Show a '1' on the LCD.
    temp -= 100;                        // Adjust the temperature variable.
#if (USE_SERIAL)
    putchar('1');                       // Output a '1' to the serial port.
#endif
  }

  LCD3 = getLCDDigit(temp / 10);        // show the 10's place on the LCD
  LCD2 = getLCDDigit(temp % 10);        // show the 1's place on the LCD

#if (USE_SERIAL)
  putchar((temp / 10) + '0');           // Output the 10's place to the serial port.
  putchar((temp % 10) + '0');           // Output the 1's place to the serial port.
  
  if(celsius != 0)
  {
    printf(" degrees Celsius.   \r"); // Output the units on the serial port.
  }
  else
  {
    printf(" degrees Fahrenheit.\r"); // Output the units on the serial port.
  }
#endif
}

/******************************************************************************
 * Get an A->D reading, convert it to a temperature, and display it.  This 
 * function is called under the timer interrupt.
 */
void readTemp()
{
  int temp;
  int adc = getADCReading();            // Read the thermistor value.
  last_adc = adc;
  temp = convertToTemp(adc);        // Convert this value to degrees Celsius.
  last_temp = temp;
  showTemp(temp);                       // Show the result.
}

/******************************************************************************
 * Initialize the timer.  With a 13.5 MHz crystal, we can get a 0.5 second 
 * interrupt period if the timer is run at clock/128...
 *
 *         timerticks * 128 = seconds * frequency
 *         timerticks = 0.5 * 13500000 / 128 ~= 52734
 *
 * Since our timers tick upwards, we want our reload value to be:
 *         reload = 64K - timerticks = 65536 - 52734 = 0x3203
 */
void initTimer()
{
  T2V0 = 0x00000;                         // Set current timer value.
  T2R0 = 0x03203;                         // Set reload value.
  T2C0 = 0x00000;                         // Set compare value to below the reload so it never generates an interrupt.

  T2CFG0_bit.T2DIV = 7;                   // Set div 128 mode.

  T2CNA0_bit.TR2 = 1;                     // Start the timer.
  T2CNA0_bit.T2POL0 = 1;                  // Set polarity for T2OE0 to high.
  T2CNA0_bit.ET2 = 1;                     // Enable interrupts.

  T2CNB0_bit.T2OE1 = 1;                   // Enable timer output.

//  IMR |= 0x08;                          // Enable the interrupts for module 3.
  asm("move M8[6].3, #1");
}

/******************************************************************************
 * Setup the external interrupts for EI 10 and 15 (pins P5.2 and P7.1).
 */
void initPushButtons()
{
    EIE1 = 0x84;                          // enable interrupts 10 and 15
    EIES1 = 0x84;                         // set up edge transitions
//    IMR |= 0x02;
  asm("move M8[6].1, #1");                // turn on interrupts for module 1
}

/******************************************************************************
 * The interrupt handler for module 1, including my external interrupts.
 * Must be careful in here to make sure I react only to the interrupts I 
 * want, since some pins are dual purpose.
 */
#pragma vector = 1
__interrupt void pushButtonInterrupt()
{
  pbcount++;                              // increment our debug variable 'pushbuttoncount'
  
  lastEIFvalue = EIF1;
  // check for interrupt 10 (0x04) or interrupt 15 (0x80)
  if (EIF1 & 0x04)
  {
      // interrupt 10
      LCFG = LCFG & 0xFD;                 // disable the LCD
      LCRA_bit.LRA++;                     // put the new resistor adjust value in there
      LCFG = LCFG | 0x02;                 // enable the LCD
  }
  if (EIF1 & 0x80)
  {
      // interrupt 15
      LCFG = LCFG & 0xFD;                 // disable the LCD
      LCRA_bit.LRA--;                     // put the new resistor adjust value in there
      LCFG = LCFG | 0x02;                 // enable the LCD
  }
  
  EIF1 = 0;                               // clear the interrupt condition
}

/******************************************************************************
 * The interrupt handler for Timer 0.  Occurs whenever the timer overflows.
 */
#pragma vector = 3
__interrupt void timerInterrupt()
{
  //Change the units occasionally.
  if(++count == 20) 
  {
    count = 0;
    celsius = ~celsius;
  }

  readTemp();           //Get and display the current temperature.
  T2CNB0_bit.TF2 = 0;   //Clear the overflow flag.
  T2CNB0_bit.TCC2 = 0;  //Clear the overflow flag.
  
  // output an 'alive' blip with the first decimal point, LCD0.7
  if (count & 1)
    LCD0 = LCD0 | 0x80;
  else
    LCD0 = LCD0 & 0x7F;
}

/******************************************************************************
 * Initialize the LCD to run off the 32 kHz, with external drive power.
 */
void initLCD()
{
  //LCRA_bit.FRM = 7;   // Set up frame frequency.
  //LCRA_bit.LCCS = 1;  // Set clock source to HF/128.
  //LCRA_bit.DUTY = 0;  // Set up static duty cycle.
  //LCRA_bit.LRA = 0;   // Set R-adj to 0.
  //LCRA_bit.LRIGC = 1; // Select external LCD drive power.
  LCRA = 0x03E0;        // Just do them all at once...

  LCFG_bit.PCF = 0x0F;// Set up all segments as outputs.
  LCFG_bit.OPM = 1;   // Set to normal operation mode.
  LCFG_bit.DPE = 1;   // Enable display.
}

/******************************************************************************
 * Main application entry point.  Initialize the serial port, timer, SPI/1407,
 * LCD, and external interrupts.  Then just do nothing.  The interrupts will
 * take care of everything else.
 */
void main()
{
#if (USE_SERIAL)
  initSerial();
#endif

  initTimer();
  init1407();
  initLCD();
  initPushButtons();

  __enable_interrupt();

  while(1);
}

